/*
 * Project name:
     ExtRes_to_AT45DB
 * Author:
     Giuseppe Finizia
     Rome - ITALY
     giuseppe.finizia@gmail.com
 * Revision History:
     1.0.0.0 - First release (19/08/2012)
 * Description:
     This project reads from a microSD card an external resources file generated by
     Visual TFT and writes data to an Atmel AT45DBxx SPI Serial Flash Memory device.
     At the moment, only AT45DB16 (2MB) and AT45DB32 (4MB) memory devices are supported
     since they have a page size of 512 bytes which is just equal to microSD sector size.
 * Development system configuration:
     Dev. Board:  EasyPIC7
                  http://www.mikroe.com/eng/products/view/757/easypic-v7-development-system/
     MCU:         PIC18F46K22 (also PIC18F45K22 is OK)
                  http://ww1.microchip.com/downloads/en/DeviceDoc/41412D.pdf
     Oscillator:  32.00000 MHz
     Modules:     Character LCD 2x16 on PORTB (RB0..RB5)
                  http://www.mikroe.com/eng/products/view/277/
                  MicroSD Card on PORTC (RC0..RC1,RC3..RC5)
                  http://www.mikroe.com/eng/products/view/369/microsd-card-board/
                  Atmel AT45DBxx SPI Serial Flash Memory on PORTC (RC2..RC5)
                  http://www.atmel.com/products/memories/sflash/dataflash.aspx
     Compiler:    mikroC PRO for PIC v5.61
                  http://www.mikroe.com/eng/products/view/7/mikroc-pro-for-pic/
 * NOTES:
     - Make sure that microSD card is properly formatted (FAT16 or just FAT) before using it
     - The name of the external resources file must be "res_file.res"
     - Set pull-down resistor on RA0 pin (input button)
     - Connect AT45DBxx 8-pin memory chip to SPI1 interface of PIC18F4xK22 on PORTC:
       - CS#  to RC2 pin      - SDI1   to RC5 pin
       - SCK1 to RC3 pin      - Reset# to Vcc
       - SDO1 to RC4 pin      - WP#    to Vcc
     - Disconnect AT45DBxx memory chip from dev. board when reprogramming PIC18F4xK22 MCU
 */

// LCD module connections
sbit LCD_D4 at LATB0_bit;
sbit LCD_D5 at LATB1_bit;
sbit LCD_D6 at LATB2_bit;
sbit LCD_D7 at LATB3_bit;
sbit LCD_RS at LATB4_bit;
sbit LCD_EN at LATB5_bit;
sbit LCD_D4_Direction at TRISB0_bit;
sbit LCD_D5_Direction at TRISB1_bit;
sbit LCD_D6_Direction at TRISB2_bit;
sbit LCD_D7_Direction at TRISB3_bit;
sbit LCD_RS_Direction at TRISB4_bit;
sbit LCD_EN_Direction at TRISB5_bit;
// End of LCD module connections

// MicroSD module connections
sbit Mmc_Chip_Select at LATC0_bit;
sbit Mmc_Chip_Select_Direction at TRISC0_bit;
sbit Mmc_Card_Detect at RC1_bit;
sbit Mmc_Card_Detect_Direction at TRISC1_bit;
// End of microSD module connections

// AT45DB memory chip connections
sbit AT45DB_Chip_Select at LATC2_bit;
sbit AT45DB_Chip_Select_Direction at TRISC2_bit;
// End of AT45DB memory chip connections

// Global variables
char memSeries;                 // Series of AT45DBxx memory device in use
unsigned long memSize;          // Size in bytes of AT45DBxx memory device in use
unsigned long resFileSize;      // Size in bytes of external resources file stored on microSD card
unsigned short SPI_StatusReg1;  // First byte of AT45DBxx1x Status Register
unsigned short SPI_StatusReg2;  // Second byte of AT45DBxx1E Status Register (only for E-series devices)
char readBuffer[512];           // Buffer containing data read from microSD card
// End of global variables

// Text strings are stored in MCU ROM area to save MCU RAM space
const char txt_01[] = "Insert microSD  ";
const char txt_02[] = "MicroSD inserted";
const char txt_03[] = "MicroSD error!  ";
const char txt_04[] = "FAT16 not found ";
const char txt_05[] = "Reinsert & reset";
const char txt_06[] = "Memory not found";
const char txt_07[] = "Turn off & check";
const char txt_08[] = "AT45DBxx1D found";
const char txt_09[] = "AT45DBxx1E found";
const char txt_10[] = "No supported mem";
const char txt_11[] = "device connected";
const char txt_12[] = "Not enough space";
const char txt_13[] = "to store Ext Res";
const char txt_14[] = "Press RA0 button";
const char txt_15[] = "to start writing";
const char txt_16[] = "Page size is now";
const char txt_17[] = "set to 512 bytes";
const char txt_18[] = "Writing data to ";
const char txt_19[] = "AT45DB memory...";
const char txt_20[] = "Error writing to";
const char txt_21[] = "45DBxx1E memory.";
const char txt_22[] = "Data written to ";
const char txt_23[] = "AT45DB memory.  ";

// Endless loop to stop code execution
void Terminate() {
  Mmc_Fat_Close();  // Close currently opened file
  while (1) {
    Delay_ms(1000);
  }
}

// Copy text string from ROM to RAM (ROM is used to save RAM space)
char* CodeTxtToRamTxt(const char* ctxt) {
  static char ramTxt[17];
  unsigned short i;  // Local counter
  for (i = 0; ramTxt[i] = ctxt[i]; i++);
  return ramTxt;
}

// Wait for microSD card insertion
void Wait_for_Card() {
  Mmc_Card_Detect_Direction = 1;  // Set microSD card detection pin to be input

  if (Mmc_Card_Detect) {
    Lcd_Cmd(_LCD_CLEAR);                   // Clear display
    Lcd_Out(1,1,CodeTxtToRamTxt(txt_01));  // Write text in first row ("Insert microSD  ")
    while (Mmc_Card_Detect) {              // Loop until card is detected
      Delay_ms(500);
    }
    Lcd_Out(1,1,CodeTxtToRamTxt(txt_02));  // Write text in first row ("MicroSD inserted")
    Delay_ms(2000);
  }
}

// Initialize SPI1 module and microSD card
void Card_Init() {
  // Initialize SPI1 module at low speed
  SPI1_Init_Advanced(_SPI_MASTER_OSC_DIV64, _SPI_DATA_SAMPLE_MIDDLE, _SPI_CLK_IDLE_LOW, _SPI_LOW_2_HIGH);

  switch (Mmc_Fat_Init()) {  // Initialize microSD card
    case 1:    // FAT16 boot sector not found
      Lcd_Out(1,1,CodeTxtToRamTxt(txt_03));  // Write text in first row ("MicroSD error!  ")
      Lcd_Out(2,1,CodeTxtToRamTxt(txt_04));  // Write text in second row ("FAT16 not found ")
      Terminate();
    case 255:  // MicroSD card not detected
      Lcd_Out(1,1,CodeTxtToRamTxt(txt_03));  // Write text in first row ("MicroSD error!  ")
      Lcd_Out(2,1,CodeTxtToRamTxt(txt_05));  // Write text in second row ("Reinsert & reset")
      Terminate();
  }

  // Reinitialize SPI1 module at higher speed
  SPI1_Init_Advanced(_SPI_MASTER_OSC_DIV4, _SPI_DATA_SAMPLE_MIDDLE, _SPI_CLK_IDLE_LOW, _SPI_LOW_2_HIGH);

  Mmc_Fat_Assign("res_file.res", 0);
  Mmc_Fat_Reset(&resFileSize);  // To read file, procedure returns size of file
}

void AT45DB_ReadStatusReg() {
  AT45DB_Chip_Select = 0;            // Select AT45DB memory chip
  SPI1_Write(0xD7);                  // Request for Status Register bytes
  SPI_StatusReg1 = SPI1_Read(0xFF);  // Read byte 1 of Status Register
  SPI_StatusReg2 = SPI1_Read(0xFF);  // Read byte 2 of Status Register
  AT45DB_Chip_Select = 1;            // Deselect AT45DB memory chip
}

void AT45DB_Check() {
  unsigned short density;  // Density Code of AT45DB Serial Flash Memory device

  AT45DB_Chip_Select = 1;            // Deselect AT45DB memory chip
  AT45DB_Chip_Select_Direction = 0;  // Set CS# pin as output

  AT45DB_ReadStatusReg();  // Read AT45DB Status Register
  if (SPI_StatusReg1 == 0x00 || SPI_StatusReg1 == 0xFF) {
      Lcd_Out(1,1,CodeTxtToRamTxt(txt_06));  // Write text in first row ("Memory not found")
      Lcd_Out(2,1,CodeTxtToRamTxt(txt_07));  // Write text in second row ("Turn off & check")
      Terminate();
  }
  if (SPI_StatusReg1 == SPI_StatusReg2) {
    memSeries = 'D';  // Atmel AT45DBxx1D detected
    Lcd_Out(1,1,CodeTxtToRamTxt(txt_08));  // Write text in first row ("AT45DBxx1D found")
  } else {
    memSeries = 'E';  // Atmel AT45DBxx1E detected
    Lcd_Out(1,1,CodeTxtToRamTxt(txt_09));  // Write text in first row ("AT45DBxx1E found")
  }
  Delay_ms(3000);

  density = SPI_StatusReg1 & 0b00111100;
  switch (density) {
    case 0b00101100:    // Atmel AT45DB161x detected
      memSize = 2097152;
      break;
    case 0b00110100:    // Atmel AT45DB321x detected
      memSize = 4194304;
      break;
    default:
      Lcd_Out(1,1,CodeTxtToRamTxt(txt_10));  // Write text in first row ("No supported mem")
      Lcd_Out(2,1,CodeTxtToRamTxt(txt_11));  // Write text in second row ("device connected")
      Terminate();
  }

  if (memSize < resFileSize) {
    Lcd_Out(1,1,CodeTxtToRamTxt(txt_12));  // Write text in first row ("Not enough space")
    Lcd_Out(2,1,CodeTxtToRamTxt(txt_13));  // Write text in second row ("to store Ext Res")
    Terminate();
  }
}

// Write data to Atmel 45DB16 Serial Flash memory
void AT45DB_Write() {
  unsigned short hiAddr;   // High byte of address
  unsigned short medAddr;  // Medium byte of address
  unsigned long tempAddr;
  unsigned long i;         // Local counter
  unsigned long cycles;
  unsigned int j;          // Local counter
  unsigned int remainder;
  char character;

  Lcd_Cmd(_LCD_CLEAR);                   // Clear display
  Lcd_Cmd(_LCD_CURSOR_OFF);              // Cursor off
  Lcd_Out(1,1,CodeTxtToRamTxt(txt_14));  // Write text in first row ("Press RA0 button")
  Lcd_Out(2,1,CodeTxtToRamTxt(txt_15));  // Write text in second row ("to start writing")

  do {  // Waiting for pressing RA0 button
    if (Button(&PORTA, 0, 1, 1))
      break;
  } while (1);

  AT45DB_ReadStatusReg();  // Read AT45DB Status Register

  // AT45DB page size must be set to 512 bytes
  if ((SPI_StatusReg1 & 0b00000001) == 0) {
    AT45DB_Chip_Select = 0;                // Select AT45DB memory chip
    SPI1_Write(0x3D);                      // Set page size to 512 bytes (opCode 1 of 4)
    SPI1_Write(0x2A);                      // Set page size to 512 bytes (opCode 2 of 4)
    SPI1_Write(0x80);                      // Set page size to 512 bytes (opCode 3 of 4)
    SPI1_Write(0xA6);                      // Set page size to 512 bytes (opCode 4 of 4)
    AT45DB_Chip_Select = 1;                // Deselect AT45DB memory chip
    Lcd_Out(1,1,CodeTxtToRamTxt(txt_16));  // Write text in first row ("Page size is now")
    Lcd_Out(2,1,CodeTxtToRamTxt(txt_17));  // Write text in second row ("set to 512 bytes")
    Delay_ms(3000);
  }

  Lcd_Out(1,1,CodeTxtToRamTxt(txt_18));  // Write text in first row ("Writing data to ")
  Lcd_Out(2,1,CodeTxtToRamTxt(txt_19));  // Write text in second row ("45DB16 memory...")

  cycles = resFileSize / 512;
  remainder = resFileSize % 512;

  for (i = 0; i < cycles; i++) {
    for (j = 0; j < 512; j++) {
      Mmc_Fat_Read(&character);
      readBuffer[j]=character;
    }

    tempAddr = i << 9;
    hiAddr = (unsigned short) (tempAddr >> 16);
    medAddr = (unsigned short) ((tempAddr & 0xFF00) >> 8);

    do {
      AT45DB_ReadStatusReg();                      // Read AT45DB Status Register
    } while ((SPI_StatusReg1 & 0b10000000) == 0);  // while "device is busy with an internal operation"

    if (memSeries == 'E')
      if ((SPI_StatusReg2 & 0b00100000) != 0) {  // if "erase or program error detected"
        Lcd_Out(1,1,CodeTxtToRamTxt(txt_20));    // Write text in first row ("Error writing to")
        Lcd_Out(2,1,CodeTxtToRamTxt(txt_21));    // Write text in second row ("45DB161E memory.")
        Terminate();
      }

    AT45DB_Chip_Select = 0;  // Select AT45DB Serial Flash chip
    SPI1_Write(0x82);        // Main Memory Page Program through Buffer with Built-In Erase
    SPI1_Write(hiAddr);
    SPI1_Write(medAddr);
    SPI1_Write(0x00);
    for (j = 0; j < 512; j++)
      SPI1_Write(readBuffer[j]);
    AT45DB_Chip_Select = 1;  // Deselect AT45DB Serial Flash chip
  }

  for (j = 0; j < remainder; j++) {
    Mmc_Fat_Read(&character);
    readBuffer[j]=character;
  }

  tempAddr = i << 9;
  hiAddr = (unsigned short) (tempAddr >> 16);
  medAddr = (unsigned short) ((tempAddr & 0xFF00) >> 8);

  do {
    AT45DB_ReadStatusReg();                      // Read AT45DB Status Register
  } while ((SPI_StatusReg1 & 0b10000000) == 0);  // while "device is busy with an internal operation"

  if (memSeries == 'E')
    if ((SPI_StatusReg2 & 0b00100000) != 0) {  // if "erase or program error detected"
      Lcd_Out(1,1,CodeTxtToRamTxt(txt_20));    // Write text in first row ("Error writing to")
      Lcd_Out(2,1,CodeTxtToRamTxt(txt_21));    // Write text in second row ("45DB161E memory.")
      Terminate();
    }

  AT45DB_Chip_Select = 0;  // Select AT45DB Serial Flash chip
  SPI1_Write(0x82);        // Main Memory Page Program through Buffer with Built-In Erase
  SPI1_Write(hiAddr);
  SPI1_Write(medAddr);
  SPI1_Write(0x00);
  for (j = 0; j < remainder; j++)
    SPI1_Write(readBuffer[j]);
  for (j = remainder; j < 512; j++)
    SPI1_Write(0x00);
  AT45DB_Chip_Select = 1;  // Deselect AT45DB Serial Flash chip

  Lcd_Out(1,1,CodeTxtToRamTxt(txt_22));  // Write text in first row ("Data written to ")
  Lcd_Out(2,1,CodeTxtToRamTxt(txt_23));  // Write text in second row ("AT45DB memory.  ")
}

// Main function
void main() {
  ANSELA = 0;       // Configure PORTA pins as digital (input buttons)
  ANSELB = 0;       // Configure PORTB pins as digital (LCD display)
  ANSELC = 0;       // Configure PORTC pins as digital (microSD card and AT45DB device)
  SLRCON = 0;       // Set output slew rate on all ports at standard rate
  TRISA = 0x01;     // Set direction of RA0 pin to be input

  Lcd_Init();       // Initialize LCD module
  Wait_for_Card();  // Wait for microSD card insertion
  Card_Init();      // Initialize SPI1 module and microSD card
  AT45DB_Check();   // Check for valid AT45DB device connected
  AT45DB_Write();   // Write data to AT45DB SPI Serial Flash memory
}